/*
** Copyright (C) Red Hat, Inc. 2005-2011
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING. If not, write to the
** Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
** MA 02139, USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#include "Communicator.h"
#include "array_auto_ptr.h"
#include "Logger.h"
#include "Time.h"

#include <sys/poll.h>
#include <stdlib.h>
#include <errno.h>

typedef struct pollfd poll_fd;

#include <map>
#include <algorithm>

using namespace ClusterMonitoring;
using namespace std;

CommDP::CommDP()
{}

CommDP::~CommDP()
{}

Communicator::Communicator(unsigned short port, CommDP& delivery_point) :
	_port(port),
	_serv_sock(_port),
	_delivery_point(delivery_point)
{
	_serv_sock.nonblocking(true);
	_connect_time = time_sec();
	_rand_state = time_mil();
	log(String("Communicator created, port ") + _port, LogCommunicator);
}

Communicator::~Communicator()
{
	stop();
	log("Communicator deleted", LogCommunicator);
}

void
Communicator::update(const String& self, const std::vector<String>& hosts, const String& msg)
{
	MutexLocker l(_mutex);

	_my_hostname = self;

	_peer_hostnames.clear();
	for (unsigned int i = 0 ; i < hosts.size() ; i++)
		_peer_hostnames.push_back(hosts[i]);

	// mark peer queues for pruning with the update from monitor (every ~5 s.)
	_out_q.push_back(msg);
	_prune_peer_queues = true;
}

void
Communicator::run()
{
	while (!shouldStop()) {
		vector<String> names;
		vector<String> que;
		String my_hostname;

		{
			MutexLocker l(_mutex);
			my_hostname = _my_hostname;

			for (unsigned int i = 0 ; i < _peer_hostnames.size() ; i++)
				names.push_back(_peer_hostnames[i]);

			for (unsigned int i = 0 ; i < _out_q.size() ; i++)
				que.push_back(_out_q[i]);
			_out_q.clear();
		}

		// remove non-peers
		vector<String> remove_us;
		for (map<String, Peer>::iterator
				iter = _peers.begin() ;
				iter != _peers.end() ;
				iter++)
		{
			const String& name = iter->first;
			if (find(names.begin(), names.end(), name) == names.end())
				remove_us.push_back(name);
		}

		for (vector<String>::iterator
				iter = remove_us.begin() ;
				iter != remove_us.end() ;
				iter++)
		{
			log("dropping connection with " + *iter, LogCommunicator);
			_peers.erase(*iter);
		}

		// connect to peers
		if (time_to_connect()) {
			for (vector<String>::iterator
					iter = names.begin() ;
					iter != names.end() ;
					iter++)
			{
				const String& name = *iter;
				if (_peers.find(name) == _peers.end()) {
					try {
						_peers.insert(pair<String, Peer>(name, Peer(name, _port)));
						log("connected to " + name + ", socket "
							+ _peers[name].get_sock_fd(), LogCommunicator);
					} catch (String e) {
						log("Communicator: connect to peers: " + e, LogCommunicator);
					} catch ( ... ) {
						log("Communicator: connect to peers: unknown error", LogCommunicator);
					}
				}
			}
		}

		// buffer msgs
		for (vector<String>::iterator iter_q = que.begin();
				iter_q != que.end();
				iter_q++)
		{
			String& msg = *iter_q;
			_delivery_point.msg_arrived(my_hostname, msg);
			for (map<String, Peer>::iterator iter_p = _peers.begin();
					iter_p != _peers.end();
					iter_p++)
			{
				iter_p->second.append(msg, _prune_peer_queues);
			}
			// clear this so we won't drop just appended items
			// (unless monitor iteration is done in between)
			// note: not locked (here+above) as synchronization is not so important
			_prune_peer_queues = false;
		}

		serve_sockets(names);
	}
}

void
Communicator::serve_sockets(const vector<String>& names)
{
	map<int, Peer> fd_peer;
	for (map<String, Peer>::iterator
		iter = _peers.begin();
		iter != _peers.end();
		iter++)
	{
		Peer& peer = iter->second;
		fd_peer.insert(pair<int, Peer>(peer.get_sock_fd(), peer));
	}

	unsigned int socks_num = fd_peer.size() + 1;

	// prepare poll structs
	array_auto_ptr<poll_fd> poll_data(new poll_fd[socks_num]);
	poll_data[0].fd = _serv_sock.get_sock();
	poll_data[0].events = POLLIN;
	poll_data[0].revents = 0;
	map<int, Peer>::iterator iter = fd_peer.begin();

	for (unsigned int i = 1 ; i < socks_num ; i++, iter++) {
		poll_data[i].fd = iter->first;
		poll_data[i].events = POLLIN;
		// enable capturing also POLLOUT when there is something to send
		if (!iter->second.outq_empty())
			poll_data[i].events |= POLLOUT;
		poll_data[i].revents = 0;
	}

	// wait for events
	int ret = poll(poll_data.get(), socks_num, 500);
	if (ret == 0)
		return;
	else if (ret == -1) {
		if (errno == EINTR)
			return;
		else
			throw String("Communicator::run(): poll() error: ")
					+ String(strerror(errno));
	}

	// handle events
	for (unsigned int i = 0 ; i < socks_num && !shouldStop() ; i++) {
		poll_fd& poll_info = poll_data[i];

		if (i == 0)  /* alt.: poll_info.fd == _serv_sock.get_sock() */
			handle_server_socket(names, poll_info.revents);
		else
			handle_client_socket(fd_peer[poll_info.fd], poll_info.revents);
	}
}

inline void
Communicator::handle_server_socket(const vector<String>& names, short revents)
{
	if (revents & POLLIN) {
		try {
			ClientSocket sock = _serv_sock.accept();
			String hostname;

			for (vector<String>::const_iterator
					iter = names.begin() ;
					iter != names.end() ;
					iter++)
			{
				const String& name = *iter;
				if (sock.connected_to(name)) {
					hostname = name;
					break;
				}
			}

			if (hostname.size()) {
				_peers.insert(pair<String, Peer>(hostname, Peer(hostname, sock)));
				log("accepted connection from " + hostname + ", socket "
					+ sock.get_sock(), LogCommunicator);
			}
		} catch (String e) {
			log("error handling server socket :" + e,
				LogCommunicator);
		} catch ( ... ) {}
	}
	if (revents & (POLLERR | POLLHUP | POLLNVAL))
		throw String("Communicator::run(): server socket error");
}

inline void
Communicator::handle_client_socket(Peer& peer, short revents)
{
	if (revents & (POLLERR | POLLNVAL)) {
		/* error condition on the socket */
		_peers.erase(peer.hostname());
		return;
	}
	if (revents & POLLIN) {
		/* try to read all available */
		vector<String> msgs;
		try {
			msgs = peer.receive();
		} catch (String e) {
			log("error receiving data from "
					+ peer.hostname() + ": " + e,
				LogCommunicator);
			_peers.erase(peer.hostname());
			return;
		} catch ( ... ) {
			log("error receiving data from " + peer.hostname(),
				LogCommunicator);
			_peers.erase(peer.hostname());
			return;
		}
#define OPTIMIZE_MSG_ARRIVED 1
#if OPTIMIZE_MSG_ARRIVED
		// optimization: skip unnecessary refreshing of the monitor's
		//               per-peer cache
		// side-effect:  skipped XML cluster updates are not checked
		//               for validity (does not change the things much)
		if (msgs.size() > 0)
			_delivery_point.msg_arrived(peer.hostname(), msgs.back());
#else
		for (unsigned int i = 0 ; i < msgs.size() ; i++)
			_delivery_point.msg_arrived(peer.hostname(), msgs[i]);
#endif
	}
	if (revents & POLLOUT) {
		/* try to write a single piece (only) of XML update */
		try {
			peer.send();
		} catch (String e) {
			log("error sending data to "
					+ peer.hostname() + " (" + this->_my_hostname + ") : " + e,
				LogCommunicator);
			_peers.erase(peer.hostname());
		} catch ( ... ) {
			log("error sending data to " + peer.hostname(),
				LogCommunicator);
			_peers.erase(peer.hostname());
		}
	}
	if (revents & POLLHUP) {
		/* handle remote peer closing the socket */
		_peers.erase(peer.hostname());
		return;
	}
}

bool
Communicator::time_to_connect()
{
	if (time_sec() > _connect_time) {
		_connect_time = time_sec() + rand_r(&_rand_state) % 15;
		return true;
	}
	return false;
}
